/* EsseDue - Semplice-text-editor */

using GLib;
using Gtk;
using GConf;
using Pango;

public class Document : Gtk.ScrolledWindow
{
		// Attributi del documento
	public string filename = "";	// Nome del file
	public bool modified = false;	// flag che indica se il documento è stato modificato
	public Gtk.TextView textview;	// la view del documento, collegata alla tab corrispondente
	public Gtk.TextBuffer buffer;	// il buffer di testo del documento
	public Gtk.Label label;			// il titolo della tab
	
	construct {
			// Inizializza la view, imposta il font e abilita la visualizzazione
		this.textview = new Gtk.TextView();
		
		string font;
		try {
			font = GConf.Client.get_default().get_string("/desktop/gnome/interface/monospace_font_name");
		} catch (GLib.Error e) {
			// TODO: aggiungere gestione errore
		}
		this.textview.modify_font(Pango.FontDescription.from_string(font));
		
		this.buffer = this.textview.get_buffer();
		this.add(this.textview);
		this.show_all();
	}
	
		// Carica un file dentro il tab. Se il file non esiste o non è leggibile, torna false
	public bool set_file(string filename)
	{
		string contents = "";
		
		if (filename != "" && !FileUtils.get_contents(filename, out contents)) {
			return false;
		}
		
		this.buffer.set_text(contents, -1);
		this.filename = filename;
		this.buffer.modified = 0;
		this.label.set_label(Path.get_basename(filename));
		this.label.set_tooltip_text(filename);
		return true;
	}
	
}

public class Mainwindow : Gtk.Window {

	private Gtk.Notebook notebook;
	
	private const string program_name = "EsseDue";
	private const string version = "1.0";
	private const string copyright = "Alessandro Pellizzari <alex@amiran.it>";
	private const string license = "This software is licensed under the terms of GNU GPL2";
	
	private const string ui = """
	<ui>
		<menubar name="MenuBar">
			<menu action="file">
				<menuitem action="file_new" />
				<menuitem action="file_open" />
				<menuitem action="file_save" />
				<menuitem action="file_saveas" />
				<separator />
				<menuitem action="file_quit" />
			</menu>
			
			<menu action="Edit">
				<menuitem action="edit_cut" />
				<menuitem action="edit_copy" />
				<menuitem action="edit_paste" />
				<menuitem action="edit_delete" />
			</menu>
			
			<menu action="View">
			</menu>
			
			<menu action="Help">
				<menuitem action="help_info" />
			</menu>
		
		</menubar>
		
		<toolbar name="ToolBar">
			<toolitem action="file_new" />
			<toolitem action="file_open" />
			<toolitem action="file_save" />
			<separator />
			<toolitem action="edit_cut" />
			<toolitem action="edit_copy" />
			<toolitem action="edit_paste" />
		</toolbar>
	</ui>
	""";
	
	
	construct
	{
		this.set_title("EsseDue");
		this.resize(640, 480);
		this.delete_event += this.on_quit;

		var v_main = new Gtk.VBox(false, 0);
		this.add(v_main);
		
		var uimanager = new UIManager();
		var accelgroup = uimanager.get_accel_group();
		this.add_accel_group(accelgroup);
		try {
			uimanager.add_ui_from_string(this.ui, -1);
		} catch (GLib.Error e) {
			// TODO: aggiungere gestione errore
		}
		
		var actiongroup = new Gtk.ActionGroup("Main");
		
			// Menu File
		actiongroup.add_action( new Action("file", "_File", null, null) );
		
		var file_new = new Action("file_new", null, null, "gtk-new");
		file_new.activate += this.on_file_new;
		actiongroup.add_action_with_accel( file_new , "<ctrl>n");
		
		var file_open = new Action("file_open", null, null, "gtk-open");
		file_open.activate += this.on_file_open;
		actiongroup.add_action_with_accel( file_open , "<ctrl>o");
		
		var file_save = new Action("file_save", null, null, "gtk-save");
		file_save.activate += this.on_file_save;
		actiongroup.add_action_with_accel( file_save , "<ctrl>s");
		
		var file_saveas = new Action("file_saveas", null, null, "gtk-save-as");
		file_saveas.activate += this.on_file_saveas;
		actiongroup.add_action( file_saveas );

		var file_quit = new Action("file_quit", null, null, "gtk-quit");
		file_quit.activate += this.on_file_quit;
		actiongroup.add_action_with_accel( file_quit , "<ctrl>q");

			// Menu Edit
		actiongroup.add_action( new Action("Edit", "_Edit", null, null) );

		var edit_cut = new Action("edit_cut", null, null, "gtk-cut");
		edit_cut.activate += this.on_edit_cut;
		actiongroup.add_action_with_accel( edit_cut , "<ctrl>x");

		var edit_copy = new Action("edit_copy", null, null, "gtk-copy");
		edit_copy.activate += this.on_edit_copy;
		actiongroup.add_action_with_accel( edit_copy , "<ctrl>c");

		var edit_paste = new Action("edit_paste", null, null, "gtk-paste");
		edit_paste.activate += this.on_edit_paste;
		actiongroup.add_action_with_accel( edit_paste , "<ctrl>v");

		var edit_delete = new Action("edit_delete", null, null, "gtk-delete");
		edit_delete.activate += this.on_edit_delete;
		actiongroup.add_action( edit_delete );
		
			// Menu View
		actiongroup.add_action( new Action("View", "_View", null, null) );
		
		
			// Menu Help
		actiongroup.add_action( new Action("Help", "_Help", null, null) );

		var help_info = new Action("help_info", null, null, "gtk-about");
		help_info.activate += this.on_help_info;
		actiongroup.add_action( help_info );
		
			// Collega le action
		uimanager.insert_action_group(actiongroup, 0);
		
		Gtk.MenuBar mb_main = (Gtk.MenuBar)uimanager.get_widget("/MenuBar");
		v_main.pack_start(mb_main, false, true, 0);
		
		Gtk.Toolbar tb_main = (Gtk.Toolbar)uimanager.get_widget("/ToolBar");
		v_main.pack_start(tb_main, false, true, 0);
		
		this.notebook = new Gtk.Notebook();
		v_main.pack_start(this.notebook, true, true, 0);
		
		var sb_main = new Gtk.Statusbar();
		v_main.pack_start(sb_main, false, true, 0);
		
		this.show_all();
		
	}
	
	private void alert(string title="", string msg="")
	{
		var req = new Gtk.Dialog.with_buttons(title,
			(Gtk.Window) this,
			Gtk.DialogFlags.MODAL,
			Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT,
			null);
		
		var label = new Gtk.Label(msg);
		label.show();
		req.vbox.pack_start(label, true, true, 10);
		
		int response = req.run();
		req.close();
	}
	
	public void add_notebook_page(string filename="")
	{
			// Crea una nuova tab, ci associa una label standard per il titolo,
			// con il pulsante di chiusura, abbina le callback e abilita la
			// visualizzazione.
			// Se viene passato un nome di file, lo apre e lo inserisce nella tab
		if (filename != "" && !FileUtils.test(filename, GLib.FileTest.IS_REGULAR)) {
			string labeltxt = "Unable to open " + filename;
			this.alert("Unable to open file", labeltxt);
			return;
		}

		var doc = new Document();
		
		var buffer = (Gtk.TextBuffer) doc.buffer;
		buffer.set_data("document", doc);
		buffer.modified_changed += this.on_document_modified;
		
		var hb = new Gtk.HBox(false, 0);
		var label = new Gtk.Label("New Document");
		doc.label = label;
		hb.pack_start( label , true, true, 0);
		
		var close = new Gtk.Button();
		close.relief = Gtk.ReliefStyle.NONE;
		close.set_image( new Gtk.Image.from_stock(Gtk.STOCK_CLOSE, Gtk.IconSize.MENU) );
		close.set_data("document", doc);
		
		close.clicked += this.on_tab_close;
		
		hb.pack_start( close , false, true, 0);
		hb.show_all();
		
		if (filename != "" && !doc.set_file(filename) ) {
			string labeltxt = "Unable to open " + filename;
			this.alert("Unable to open file", labeltxt);
		}
		
		this.notebook.append_page( doc, hb );
		
		// Corretto il bug nelle vapi di vala >0.5.2
		//this.notebook.set_current_page( NotebookPage.num(this.notebook, doc) );
		
		this.notebook.set_current_page( this.notebook.get_current_page());

	}
	
	private Document get_active_document()
	{
		return (Document)this.notebook.get_nth_page(this.notebook.get_current_page());
	}


		/**************
		 * 	Callbacks *
		 **************/	
	
	[CCode (instance_pos = -1)]
	public bool on_quit(Mainwindow w, Gdk.Event e)
	{
		if (this.check_quit()) {
			Gtk.main_quit();
			return false;
		} else {
			return true;
		}
	}

	[CCode (instance_pos = -1)]
	public void on_file_quit(Gtk.Action? a)
	{
		if (this.check_quit()) {
			Gtk.main_quit();
		}
	}
	
	private bool check_quit()
	{
		int n_pages = this.notebook.get_n_pages();
		for (int i=0; i<n_pages; i++) {
			Document doc = (Document)this.notebook.get_nth_page(i);
			if ( doc.buffer.get_modified() ) {
				var req = new Gtk.Dialog.with_buttons("Document close",
					(Gtk.Window) this,
					Gtk.DialogFlags.MODAL,
					Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
					Gtk.STOCK_YES, Gtk.ResponseType.ACCEPT,
					null);
				
				var label = new Gtk.Label("There are some unsaved documents. Are you sure you want to quit?");
				label.show();
				req.vbox.pack_start(label, true, true, 10);
				
				int response = req.run();
				req.close();
				
				if (response == Gtk.ResponseType.ACCEPT) {
					return true;
				} else {
					return false;
				}
			}
		}
		return true;
	}
	
	/**
	 * Se il documento viene modificato, cambia il titolo della tab per
	 * visualizzare un asterisco
	 */
	[CCode (instance_pos = -1)]
	public void on_document_modified(Gtk.TextBuffer widget)
	{
		if (widget.get_modified()) {
			var doc = (Document) (widget.get_data("document"));
			string title = doc.label.get_label();
			
			if ( ! title.has_prefix("* ")) {
				title = "* "+title;
				doc.label.set_label(title);
			}
		}
	}
	
	/**
	 * Se una tab viene chiusa, controlla se il documento è stato modificato,
	 * e in caso presenta un requester per salvarlo o annullare la chiusura
	 */
	[CCode (instance_pos = -1)]
	public void on_tab_close(Gtk.Button widget)
	{
		var sw = (Document) widget.get_data("document");
		
		// Corretto il bug nelle vapi di vala >0.5.2
		//int page = NotebookPage.num(this.notebook,sw);
		int page = this.notebook.get_current_page();
		
		if (sw.buffer.get_modified()) {
			
			// La imposta come pagina attiva
			this.notebook.set_current_page(page);
			
			var req = new Gtk.Dialog.with_buttons("Document close",
						(Gtk.Window) this,
						Gtk.DialogFlags.MODAL,
						Gtk.STOCK_SAVE, Gtk.ResponseType.ACCEPT,
						Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
						Gtk.STOCK_OK, Gtk.ResponseType.OK,
						null);
			
			var label = new Gtk.Label("Are you sure you want to close this unsaved document?");
			label.show();
			req.vbox.pack_start(label, true, true, 10);
			
			int response = req.run();
			req.close();
			
			if (response == (int)Gtk.ResponseType.OK) {
				// Chiudi tutto senza salvare
				this.notebook.remove_page(page);
			} else if (response == (int) Gtk.ResponseType.ACCEPT) {
				// Salva e chiudi
				this.on_file_save( null );
				if (sw.buffer.modified == 0) {
					this.notebook.remove_page(page);
				}
			} else {
				// Non fare niente
			}
		} else {
			this.notebook.remove_page(page);
		}
	}
	
	/**
	 * Aggiunge una tab
	 */
	[CCode (instance_pos = -1)]
	public void on_file_new(Gtk.Action? a)
	{
		this.add_notebook_page();
	}

	/**
	 * Carica un file, prepara la tab per contenerlo e imposta le variabili
	 * di stato relative
	 */
	[CCode (instance_pos = -1)]
	public void on_file_open(Gtk.Action? a)
	{
		var filereq = new Gtk.FileChooserDialog("Choose file", 
						(Gtk.Window) this, Gtk.FileChooserAction.OPEN,
						Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
						Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT,
						null);
		
		if (filereq.run() == Gtk.ResponseType.ACCEPT) {
			string filename = filereq.get_filename();
			filereq.close();
			var active_doc = this.get_active_document();
			if (active_doc.filename == "") {
				if ( !FileUtils.test(filename, GLib.FileTest.IS_REGULAR) || !active_doc.set_file(filename) ) {
					string labeltxt = "Unable to open " + filename;
					this.alert("Unable to open file", labeltxt);
				}
			} else {
				this.add_notebook_page(filename);
			}
		} else {
			filereq.close();
		}
	}

	/**
	 * Salva il file, eventualmente chiedendone il nome tramite requester,
	 * e imposta lo stato relativo del documento
	 */
	[CCode (instance_pos = -1)]
	public void on_file_save(Gtk.Action? a)
	{
		// bool save = true;
		var active_doc = this.get_active_document();
		
		if (!active_doc.buffer.get_modified()) {
			return;
		}
		
		if (active_doc.filename == "") {
			this.on_file_saveas(a);
		} else {
			Gtk.TextIter s, e;
			active_doc.buffer.get_bounds(out s, out e);
			string contents = active_doc.buffer.get_text(s, e, false);
			if (!FileUtils.set_contents(active_doc.filename, contents, -1)) {
				this.alert("Errore", "Errore durante il salvataggio\n");
			}
			active_doc.buffer.modified = 0;
			active_doc.label.set_label(Path.get_basename(active_doc.filename));
			active_doc.label.set_tooltip_text(active_doc.filename);
		}
	}

	/**
	 * Presenta un requester che chiede il nome del file da salvare, imposta
	 * lo stato relativo e quindi richiama la funzione per il salvataggio
	 */
	[CCode (instance_pos = -1)]
	public void on_file_saveas(Gtk.Action? a)
	{
		string filename = this.get_filename();
		if (filename != null) {
			var active_doc = this.get_active_document();
			active_doc.filename = filename;
			this.on_file_save(a);
		}
	}
	
	/**
	 * Funzione di appoggio che apre un requester per chiedere il nome di un file
	 */
	public string get_filename()
	{
		string filename = null;
		var filereq = new Gtk.FileChooserDialog("Choose file name", 
						(Gtk.Window) this, Gtk.FileChooserAction.SAVE,
						Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
						Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT,
						null);
		
		if (filereq.run() == Gtk.ResponseType.ACCEPT) {
			filename = filereq.get_filename();
		}
		
		filereq.close();
		return filename;
	}		
	
	/**
	 * Taglia il testo selezionato dal buffer e lo salva in una clipboard
	 */
	[CCode (instance_pos = -1)]
	public void on_edit_cut(Gtk.Action? a)
	{
		var active_doc = this.get_active_document();
		var clipboard = active_doc.textview.get_clipboard(Gdk.SELECTION_CLIPBOARD);
		
		active_doc.buffer.cut_clipboard(clipboard, active_doc.textview.get_editable());
	}

	/**
	 * Copia il testo selezionato nel buffer in una clipboard
	 */
	[CCode (instance_pos = -1)]
	public void on_edit_copy(Gtk.Action? a)
	{
		var active_doc = this.get_active_document();
		var clipboard = active_doc.textview.get_clipboard(Gdk.SELECTION_CLIPBOARD);
		
		active_doc.buffer.copy_clipboard(clipboard);
	}

	/**
	 * Incolla il testo nella clipboard nel buffer alla posizione corrente
	 */
	[CCode (instance_pos = -1)]
	public void on_edit_paste(Gtk.Action? a)
	{
		var active_doc = this.get_active_document();
		var clipboard = active_doc.textview.get_clipboard(Gdk.SELECTION_CLIPBOARD);
		
		active_doc.buffer.paste_clipboard(clipboard, null, active_doc.textview.get_editable());

	}

	/**
	 * Cancella dal buffer il testo selezionato
	 */
	[CCode (instance_pos = -1)]
	public void on_edit_delete(Gtk.Action? a)
	{
		var active_doc = this.get_active_document();
		
		active_doc.buffer.delete_selection(true, active_doc.textview.get_editable());
	}

	/**
	 * Presenta un about dialog con le informazioni sul software
	 */
	[CCode (instance_pos = -1)]
	public void on_help_info(Gtk.Action? a)
	{
		var about = new Gtk.AboutDialog();
		about.program_name = this.program_name;
		about.version = this.version;
		about.copyright = this.copyright;
		about.license = this.license;
		about.run();
		about.close();
	}

}

/**
 * Classe principale del programma, il punto di entrata
 * Inizializza la GUI e fa girare il gestore di eventi di Gtk
 */
public class Esse : GLib.Object
{

	public static int main(string[] args)
	{
		
		Gtk.init(ref args);
		var mainwindow = new Mainwindow();
		
		int i=0;
		foreach (string arg in args) {
			if (i++ != 0) {
				mainwindow.add_notebook_page(arg);
			}
		}
		if (i==1) {	
			// Nessun file sulla command line, aggiunge un tab vuoto
			mainwindow.add_notebook_page();
		}
		
		Gtk.main();
		return 0;
	}
}
